<?php

namespace Drupal\quiz_item_analysis\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Form\FormState;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\paragraphs\Entity\Paragraph;
use Drupal\quiz\Entity\Quiz;
use Drupal\quiz_item_analysis\Util\QuizItemAnalysisUtil;
use Drupal\views\Views;
use Symfony\Component\HttpFoundation\Response;


/**
 * Returns responses for Quiz Item Analysis routes.
 */
class QuizItemAnalysisController extends ControllerBase {

  /**
   * Quiz analysis. Builds the response.
   *
   */
  public function quiz_item_analysis($quiz, $compare = NULL) {
    $build['quiz_content'] = [
      '#type' => 'item',
      //'#markup' => $this->t('It works!'. $quiz . $compare),
    ];

    $args = array($quiz);
    $view = Views::getView('quiz_item_analysis_result');
    $view->setDisplay('default');
    $view->initHandlers();

    $form_state = new FormState();

    $form_state->setFormState([
      'view' => $view,
      'display' => $view->display_handler->display,
      'exposed_form_plugin' => $view->display_handler->getPlugin('exposed_form'),
      'method' => 'get',
      'rerender' => TRUE,
      'no_redirect' => TRUE,
      'always_process' => TRUE,
    ]);
    $form_state->setValue(['display','id'] , 'default');
    $form_state->setValue(['display','display_plugin'] , '');
    $form_state->setValue(['display','display_title'] , '');
    $form_state->setMethod('get');

    $exposed_form = \Drupal::formBuilder()->buildForm('Drupal\views\Form\ViewsExposedForm' , $form_state);
    $exposed_form['#action'] = \Drupal::request()->getRequestUri();
    $exposed_form['first_attempt'] = array(
      '#title' => 'First attempt',
      '#type' => 'checkbox',
      '#name' => 'first_attempt',
    );
    if (!empty($_GET['first_attempt'])) {
      $exposed_form['first_attempt']['#attributes']['checked'] = 1;
    }

    $view->initDisplay();
    $view->setArguments($args);
    $view->execute();

    $build['filter_form'] = $exposed_form;

    $rids = $this->quiz_item_analysis_filter_results($view);

    if (!empty($rids)) {

      $variables = quiz_item_analysis_table($quiz, $rids);

   //   $compare = 8; // testing purpose only
      if ($compare) {

        $quiz1 = Quiz::load($quiz);
        $quiz1_title = $quiz1->get('title')->getValue()[0]['value'];
        $quiz1_questions = $quiz1->getQuestions();
        foreach ($quiz1_questions as $quizQuestionRelationship) {

          $quizQuestion = $quizQuestionRelationship->getQuestion();
          $qlist1[] = $quizQuestion->id();
        }

        $quiz2 = Quiz::load($compare);
        $quiz2_questions = $quiz2->getQuestions();
        $quiz2_title = $quiz2->get('title')->getValue()[0]['value'];
        $qlist2 = array();
        foreach ($quiz2_questions as $quizQuestionRelationship) {

          $quizQuestion2 = $quizQuestionRelationship->getQuestion();
          $qlist2[] = $quizQuestion2->id();
        }

        $diff = array_diff($qlist1, $qlist2);
        if (!empty($diff)) {
          $build['quiz_compare_msg'] = [
            '#type' => 'item',
            '#markup' => $this->t("The selected quiz cannot be compared as they do not contain the same questions."),
          ];
          return $build;
        }

        $view->destroy();
        $view->setArguments(array($compare));
        $view->execute();

        $rids = $this->quiz_item_analysis_filter_results($view);

        if(empty($rids))
        {
          $build['compare_msg'] = [
            '#type' => 'item',
            '#markup' => $this->t('Comparison quiz does not have any attempts.'),
          ];
          return $build;
        }
        $variables2 = quiz_item_analysis_table($compare, $rids);

        foreach ($variables['rows'] as $idx => &$cells) {
          foreach ($variables2['rows'][$idx] as $key => $col) {
            if (in_array($key, array(
              'rcounts',
              'rper',
              'qcount',
              'corfac',
              'stddev',
              'disc_idx',
              'disc_eff',
            ))) {
              $cells[] = $col;
            }
          }
        }
        $variables['header'] = array_merge($variables['header'], array(
          'Chosen',
          '%',
          'Total',
          'Cor. Fac.',
          'Std. Dev.',
          'Disc. Idx.',
          'Disc. Eff.',
        ));

        $topheader = array(
          array(
            'colspan' => 6,
          ),
          array(
            'data' => $quiz1_title,
            'colspan' => 7,
          ),
          array(
            'data' => $quiz2_title,
            'colspan' => 7,
          ),
        );

        foreach ($variables['header'] as &$item) {
          $newCell = array(
            'data' => $item,
            'header' => TRUE,
          );
          $item = $newCell;
        }

        array_unshift($variables['rows'], $variables['header']);
        $variables['header'] = $topheader;
      }
    }

    if (isset($_GET['csv'])) {

      if (\Drupal::moduleHandler()->moduleExists('format_csv')) {

        $file_name= "$quiz-analysis.csv";
        $csv_data = theme_format_csv($variables);

        $response = new Response();
        $response->headers->set('Content-Type', 'text/csv');
        $response->headers->set('Content-Disposition', "attachment; filename=$file_name");
        $response->setContent($csv_data);

        $build['content'] = [
          '#type' => 'item',
          '#markup' => $response->send(),
        ];
        return $build;
      }
      else {
        $build['compare_msg'] = [
          '#type' => 'item',
          '#markup' => $this->t('Please enable the format_csv module to export this report.'),
        ];
        return $build;
      }
    }
    else {
      if (!empty($variables)) {
        $params = \Drupal::request()->query->all();
        $params['csv'] = 1;
        $current_path = \Drupal::service('path.current')->getPath();

        $options = array(
          'query'  => $params,
          'html'   => TRUE,
        );
        $build['data_table'] = [
          '#type' => 'table',
          '#header' => $variables['header'],
          '#rows' => $variables['rows'],
          '#attributes' => array('style' => array('color' => 'black')),
        ];
        $internal_link = Link::fromTextAndUrl(t('CSV'), Url::fromUri('internal:'.$current_path, $options))->toString();

        $build['csv_download'] = [
          '#type' => 'item',
          '#markup' => $internal_link,
        ];
      }
      else {
        $build['result_message'] = [
          '#type' => 'item',
          '#markup' => $this->t('No results.'),
        ];

      }

      return $build;
    }

  }

  /**
   * Get a list of result IDs to process from a view.
   */
  function quiz_item_analysis_filter_results($view) {

    $rids = array();
    $users = array();

    // Check if the view is not empty and return result ids.
    if (!empty($view->result)) {

      foreach ($view->result as $id => $result) {

        // option -1
        $quiz_entity = $result->_entity;
        $uid = $quiz_entity->get('uid')->getString();
        $result_id = $quiz_entity->get('result_id')->value;
        $time_end = $quiz_entity->get('time_end')->value;
       // $score = $quiz_entity->get('score')->value;

        // option -2
       /* foreach ($view->field as $fid => $field) {
          if ($fid == 'uid')
            $uid = $field->getValue($result);
          else if ($fid == 'result_id')
            $result_id = $field->getValue($result);
          else if ($fid == 'time_end')
            $time_end = $field->getValue($result);

        }*/
        if(!empty($result_id))
          $users[$uid][$result_id] = $time_end;
      }

      foreach ($users as $uid => $results) {
        if (!empty($_GET['first_attempt'])) {
          asort($results);
          $rids[] = key($results);
        }
        else {
          $rids = array_merge($rids, array_keys($results));
        }

      }

    }

    return $rids;
  }


}

/**
 * Generate an item analysis for a single question.
 */
function quiz_item_analysis_table($quiz, $rids = array()) {
  // Get the question relationships for this quiz.

  $quiz_node = Quiz::load($quiz);
  $questions = $quiz_node->getQuestions();

  // Get the alternatives for all the questions.
  $alternatives = [];

  foreach ($questions as $quizQuestionRelationship) {
   $answer_choices = array();
    $quizQuestion = $quizQuestionRelationship->getQuestion();
    $question_id = $quizQuestion->id();

      if ($quizQuestion->bundle() == 'multichoice') {

        // Answer choices for each question
        foreach ($quizQuestion->get('alternatives')->referencedEntities() as $alternative) {
          /* @var $alternative Paragraph */
          $choice_vid = $alternative->getRevisionId();
          $answer_choices['id'] = $choice_vid;
          $answer_choices['answer']['value'] = $alternative->get('multichoice_answer')->getValue()[0]['value'];
          $answer_choices['answer']['format'] = $alternative->get('multichoice_answer')->getValue()[0]['format'];
          $answer_choices['score_if_chosen'] = $alternative->get('multichoice_score_chosen')->getString() ?: 0;
          $answer_choices['score_if_not_chosen'] = $alternative->get('multichoice_score_not_chosen')->getString();
          $alternatives[$question_id][] = $answer_choices;
        }

      }else if ($quizQuestion->bundle() == 'truefalse') {

        $answer_choices[1]['id'] = 1;
        $answer_choices[0]['id'] = 0;
        $answer_choices[1]['answer']['value'] = 'True';
        $answer_choices[1]['answer']['format'] = 'plain_text';
        $answer_choices[0]['answer']['value'] = 'False';
        $answer_choices[0]['answer']['format'] = 'plain_text';
        $correct_answer = $quizQuestion->get('truefalse_correct')->getString();
        $answer_choices[(int) $correct_answer]['score_if_chosen'] = 1;
        $answer_choices[(int) !$correct_answer]['score_if_chosen'] = 0;

        $alternatives[$question_id] = $answer_choices;

      }

    }

  $result_ids = implode(",", $rids);

  $sql= "Select question_vid, qramc.multichoice_answer_value as answer_id, is_correct, points_awarded, count(question_vid) as answer_count from {quiz_result_answer} qra
INNER JOIN {quiz_result_answer__multichoice_answer} qramc ON qra.result_answer_id = qramc.entity_id
INNER JOIN {paragraph__multichoice_answer} pmca on qramc.multichoice_answer_value = pmca.revision_id
where qra.result_id in ($result_ids)
group by question_vid, answer_id, is_correct, points_awarded
UNION
Select question_vid, qatf.truefalse_answer_value as answer_id, is_correct, points_awarded, count(question_vid) as answer_count from {quiz_result_answer} qra
INNER JOIN {quiz_result_answer__truefalse_answer} qatf ON qra.result_answer_id = qatf.revision_id
where qra.result_id in ($result_ids)
group by question_vid, answer_id, is_correct, points_awarded";


 // $resultm = \Drupal::database()->query($sql, [':rids' => $result_ids]); // query returns only 3 rows instead 8 for (rid= 8,9,10,11,12)
  $resultm = \Drupal::database()->query($sql);
  //$resultm->allowRowCount = TRUE;

  while ($row = $resultm->fetch()) {

    // Question ID
    $qid = $row->question_vid;

    // Track the points earned.
    if (!isset($qagg[$qid]['points'])) {
      $qagg[$qid]['points'] = array();
    }

    // Number of responses to this answer ID.
    $qagg[$qid]['answers_agg'][$row->answer_id] = $row->answer_count;

    // Store point values by question.
    $points = array_fill(0, $row->answer_count, $row->points_awarded);
    $qagg[$qid]['points'] = array_merge($qagg[$qid]['points'], $points);

    // Number of responses overall.
     @$qagg[$qid]['rcount'] += $row->answer_count;

    // Store how many answers were correct and also incorrect.
    if ($row->is_correct) {
      $qagg[$qid]['correct'] = $row->answer_count;
    }

  }


  // Half the results for our window reporting.
  $middle = count($rids) / 2;

  $count = 0;
  foreach ($questions as $question) {
    $count++;

    $quizQuestion = $question->getQuestion();
    $qid = $quizQuestion->id();

    if ($quizQuestion->bundle() == 'multichoice' || $quizQuestion->bundle() == 'truefalse') {

      $row['num'] = $count;
      $row['qid'] = $qid;
      $row['qtype'] = $quizQuestion->bundle();
      $row['qname'] = $quizQuestion->get('title')->value;
      $first = TRUE;

      if (is_array($alternatives[$row['qid']])) {
        foreach ($alternatives[$row['qid']] as $answer) {

          $aid = $answer['id'];
          $stats = $qagg[$qid]['answers_agg'];
          $row['atext'] = check_markup($answer['answer']['value'], $answer['answer']['format']);
          $row['partial'] = number_format($answer['score_if_chosen'], 2);
          @$row['rcounts'] = number_format($stats[$aid]);
          @$row['rper'] = number_format($stats[$aid] / $qagg[$qid]['rcount'] * 100) . '%';


      if ($first) {
        // Number of responses.
        $row['qcount'] = number_format($qagg[$qid]['rcount']);

        // Correct factor.
        if (!empty($qagg[$qid]['correct'])) {
          $row['corfac'] = number_format($qagg[$qid]['correct'] / $qagg[$qid]['rcount'] * 100) . '%';
        }
        else {
          $row['corfac'] = '';
        }

        // Standard deviation.
        $row['stddev'] = number_format(QuizItemAnalysisUtil::stats_standard_deviation($qagg[$qid]['points'] ?? []) * 100, TRUE) . '%';

        // Calculate discrimination index.
        // Create temporary table to split results in half based on score.
        $sql = "SELECT @s:=@s+1 row_number, qnr.result_id, qid, vid, uid, qnr.is_evaluated, score, result_answer_id, question_id, question_vid, is_correct, points_awarded
              FROM (select @s:=0) AS s, {quiz_result} qnr
              INNER JOIN {quiz_result_answer} qnra ON (qnr.result_id = qnra.result_id)
              WHERE qnr.qid = :quiz_id AND question_id = :qid AND qnr.result_id in ($result_ids)
              ORDER BY score DESC";
        $temp_table_name = \Drupal::database()->queryTemporary($sql, array(
          ':quiz_id' => $quiz,
          ':qid' => $qid,
          //':rids' => $rids,
        ));

        // Correct answers from highest scorers.
        $hi_correct = \Drupal::database()->query("SELECT sum(is_correct) AS correct FROM {{$temp_table_name}} WHERE row_number <= :middle", array(':middle' => $middle))->fetchField();

        // Correct answers from lowest scorers.
        $lo_correct = \Drupal::database()->query("SELECT sum(is_correct) AS correct FROM {{$temp_table_name}} WHERE row_number > :middle", array(':middle' => $middle))->fetchField();

        // Discrimination index is the number of takers who got it correct
        // in the high group minus the number correct in the low group,
        // divided by the group size.
        $disc = ($hi_correct - $lo_correct) / $middle;
        $row['disc_idx'] = number_format($disc * 100, 2) . '%';

        // Calculate covariance.
        // Get points from this question in this quiz.
        $points = \Drupal::database()->query("SELECT result_answer_id, points_awarded
              FROM {quiz_result} qr
              INNER JOIN {quiz_result_answer} qra ON (qr.result_id = qra.result_id)
              WHERE qr.qid = :quiz_id AND question_id = :qid AND qr.result_id in ($result_ids)
              ORDER BY qr.result_id", array(
          ':quiz_id' => $quiz,
          ':qid' => $qid,
          //':rids' => $rids,
        ))->fetchAllKeyed();
        $int_points = array_values(array_map('intval', $points));

        // Get scores from this question in this quiz.
        $score = \Drupal::database()->query("SELECT result_answer_id, score
              FROM {quiz_result} qr
              INNER JOIN {quiz_result_answer} qra ON (qr.result_id = qra.result_id)
              WHERE qr.qid = :quiz_id AND question_id = :qid AND qr.result_id in ($result_ids)
              ORDER BY qr.result_id", array(
          ':quiz_id' => $quiz,
          ':qid' => $qid,
         // ':rids' => $rids,
        ))->fetchAllKeyed();
        $int_scores = array_values(array_map('intval', $score));

        $covar =  QuizItemAnalysisUtil::stats_covariance($int_points, $int_scores);
        sort($int_points);
        sort($int_scores);
        $covar_max = QuizItemAnalysisUtil::stats_covariance($int_points, $int_scores);

        if ($covar_max) {
          $row['disc_eff'] = number_format(($covar / $covar_max) * 100, 2) . '%';
        }else{
          $row['disc_eff'] = '';
        }

        $first = FALSE;
      }
      else {
        $row['num'] = '';
        $row['qid'] = '';
        $row['qtype'] = '';
        $row['qname'] = '';
        $row['qcount'] = '';
        $row['corfac'] = '';
        $row['stddev'] = '';
        $row['disc_idx'] = '';
        $row['disc_eff'] = '';
      }

      $rows[] = $row;
       }
     }
    }
  }

  $variables['rows'] = $rows;
  $variables['header'] = array(
    '#',
    'ID',
    'Type',
    'Question',
    'Answer',
    'Score',
    'Chosen',
    '%',
    'Total',
    'Cor. Fac.',
    'Std. Dev.',
    'Disc. Idx.',
    'Disc. Eff.',
  );

  return $variables;


}
